@@ -33,34 +33,47 @@
 
 #include "gnc-commodity.h"
 #include "gnc-engine-util.h"
+#include "guid.h"
 
+static short module = MOD_ENGINE; 
 
 /* parts per unit is nominal, i.e. number of 'partname' units in
  * a 'unitname' unit.  fraction is transactional, i.e. how many
  * of the smallest-transactional-units of the currency are there
  * in a 'unitname' unit. */ 
 
-struct _gnc_commodity { 
-  char   * fullname;  
-  char   * namespace;
-  char   * mnemonic;
-  char   * printname;
-  char   * exchange_code;  /* CUSIP or other identifying code */
-  int    fraction;
-  char   * unique_name;  
-  gint16 mark;             /* user-defined mark, handy for traversals */
+struct gnc_commodity_s { 
+  char    * fullname;  
+  char    * namespace;
+  char    * mnemonic;
+  char    * printname;
+  char    * exchange_code;  /* CUSIP or other identifying code */
+  int       fraction;
+  char    * unique_name;  
+  gint16    mark;           /* user-defined mark, handy for traversals */
 };
 
-struct _gnc_commodity_namespace {
+struct gnc_commodity_namespace_s {
   GHashTable * table;
 };
 
-struct _gnc_commodity_table {
+struct gnc_commodity_table_s {
   GHashTable * table;
 };
 
-typedef struct _gnc_commodity_namespace gnc_commodity_namespace;
+typedef struct gnc_commodity_namespace_s gnc_commodity_namespace;
 
+struct gnc_new_iso_code {
+  const char *old_code;
+  const char *new_code;
+} gnc_new_iso_codes[] = {
+  {"RUB", "RUR"}, // Russian Ruble
+  {"PLZ", "PLN"}, // Polish Zloty
+  {"UAG", "UAH"}, // Ukraine Hryvnia
+};
+#define GNC_NEW_ISO_CODES \
+        (sizeof(gnc_new_iso_codes) / sizeof(struct gnc_new_iso_code))
+  
 /********************************************************************
  * gnc_commodity_new
  ********************************************************************/
@@ -110,15 +123,30 @@ gnc_commodity_new(const char * fullname,
  ********************************************************************/
 
 void
-gnc_commodity_destroy(gnc_commodity * cm) {
+gnc_commodity_destroy(gnc_commodity * cm)
+{
   if(!cm) return;
+
   g_free(cm->fullname);
+  cm->fullname = NULL;
+
   g_free(cm->printname);
+  cm->printname = NULL;
+
   g_free(cm->namespace);
+  cm->namespace = NULL;
+
   g_free(cm->exchange_code);
+  cm->exchange_code = NULL;
+
   g_free(cm->mnemonic);
+  cm->mnemonic = NULL;
+
   g_free(cm->unique_name);
+  cm->unique_name = NULL;
+
   cm->mark = 0;
+
   g_free(cm);
 }
 
@@ -128,7 +156,8 @@ gnc_commodity_destroy(gnc_commodity * cm) {
  ********************************************************************/
 
 const char *
-gnc_commodity_get_mnemonic(const gnc_commodity * cm) {
+gnc_commodity_get_mnemonic(const gnc_commodity * cm) 
+{
   if(!cm) return NULL;
   return cm->mnemonic;
 }
@@ -138,7 +167,8 @@ gnc_commodity_get_mnemonic(const gnc_commodity * cm) {
  ********************************************************************/
 
 const char *
-gnc_commodity_get_printname(const gnc_commodity * cm) {
+gnc_commodity_get_printname(const gnc_commodity * cm) 
+{
   if(!cm) return NULL;
   return cm->printname;
 }
@@ -149,7 +179,8 @@ gnc_commodity_get_printname(const gnc_commodity * cm) {
  ********************************************************************/
 
 const char *
-gnc_commodity_get_namespace(const gnc_commodity * cm) {
+gnc_commodity_get_namespace(const gnc_commodity * cm) 
+{
   if(!cm) return NULL;
   return cm->namespace;
 }
@@ -160,7 +191,8 @@ gnc_commodity_get_namespace(const gnc_commodity * cm) {
  ********************************************************************/
 
 const char *
-gnc_commodity_get_fullname(const gnc_commodity * cm) {
+gnc_commodity_get_fullname(const gnc_commodity * cm) 
+{
   if(!cm) return NULL;
   return cm->fullname;
 }
@@ -171,7 +203,8 @@ gnc_commodity_get_fullname(const gnc_commodity * cm) {
  ********************************************************************/
 
 const char *
-gnc_commodity_get_unique_name(const gnc_commodity * cm) {
+gnc_commodity_get_unique_name(const gnc_commodity * cm) 
+{
   if(!cm) return NULL;
   return cm->unique_name;
 }
@@ -182,7 +215,8 @@ gnc_commodity_get_unique_name(const gnc_commodity * cm) {
  ********************************************************************/
 
 const char * 
-gnc_commodity_get_exchange_code(const gnc_commodity * cm) {
+gnc_commodity_get_exchange_code(const gnc_commodity * cm) 
+{
   if(!cm) return NULL;
   return cm->exchange_code;
 }
@@ -192,7 +226,8 @@ gnc_commodity_get_exchange_code(const gnc_commodity * cm) {
  ********************************************************************/
 
 int
-gnc_commodity_get_fraction(const gnc_commodity * cm) {
+gnc_commodity_get_fraction(const gnc_commodity * cm) 
+{
   if(!cm) return 0;
   return cm->fraction;
 }
@@ -202,7 +237,8 @@ gnc_commodity_get_fraction(const gnc_commodity * cm) {
  ********************************************************************/
 
 gint16
-gnc_commodity_get_mark(const gnc_commodity * cm) {
+gnc_commodity_get_mark(const gnc_commodity * cm) 
+{
   if(!cm) return 0;
   return cm->mark;
 }
@@ -212,7 +248,8 @@ gnc_commodity_get_mark(const gnc_commodity * cm) {
  ********************************************************************/
 
 void
-gnc_commodity_set_mnemonic(gnc_commodity * cm, const char * mnemonic) {
+gnc_commodity_set_mnemonic(gnc_commodity * cm, const char * mnemonic) 
+{
   if(!cm) return;
   if(cm->mnemonic == mnemonic) return;
 
@@ -228,7 +265,8 @@ gnc_commodity_set_mnemonic(gnc_commodity * cm, const char * mnemonic) {
  ********************************************************************/
 
 void
-gnc_commodity_set_namespace(gnc_commodity * cm, const char * namespace) {
+gnc_commodity_set_namespace(gnc_commodity * cm, const char * namespace) 
+{
   if(!cm) return;
   if(cm->namespace == namespace) return;
 
@@ -244,7 +282,8 @@ gnc_commodity_set_namespace(gnc_commodity * cm, const char * namespace) {
  ********************************************************************/
 
 void
-gnc_commodity_set_fullname(gnc_commodity * cm, const char * fullname) {
+gnc_commodity_set_fullname(gnc_commodity * cm, const char * fullname) 
+{
   if(!cm) return;
   if(cm->fullname == fullname) return;
 
@@ -260,7 +299,8 @@ gnc_commodity_set_fullname(gnc_commodity * cm, const char * fullname) {
 
 void
 gnc_commodity_set_exchange_code(gnc_commodity * cm, 
-                                const char * exchange_code) {
+                                const char * exchange_code) 
+{
   if(!cm) return;
   if(cm->exchange_code == exchange_code) return;
 
@@ -273,7 +313,8 @@ gnc_commodity_set_exchange_code(gnc_commodity * cm,
  ********************************************************************/
 
 void
-gnc_commodity_set_fraction(gnc_commodity * cm, int fraction) {
+gnc_commodity_set_fraction(gnc_commodity * cm, int fraction) 
+{
   if(!cm) return;
   cm->fraction = fraction;
 }
@@ -283,7 +324,8 @@ gnc_commodity_set_fraction(gnc_commodity * cm, int fraction) {
  ********************************************************************/
 
 void
-gnc_commodity_set_mark(gnc_commodity * cm, gint16 mark) {
+gnc_commodity_set_mark(gnc_commodity * cm, gint16 mark) 
+{
   if(!cm) return;
   cm->mark = mark;
 }
@@ -294,7 +336,8 @@ gnc_commodity_set_mark(gnc_commodity * cm, gint16 mark) {
  ********************************************************************/
 
 gboolean
-gnc_commodity_equiv(const gnc_commodity * a, const gnc_commodity * b) {
+gnc_commodity_equiv(const gnc_commodity * a, const gnc_commodity * b) 
+{
   if(a == b) return TRUE;
   if(!a || !b) return FALSE;
   if(safe_strcmp(a->namespace, b->namespace) != 0) return FALSE;
@@ -302,6 +345,51 @@ gnc_commodity_equiv(const gnc_commodity * a, const gnc_commodity * b) {
   return TRUE;
 }
 
+gboolean
+gnc_commodity_equal(const gnc_commodity * a, const gnc_commodity * b)
+{
+  if (a == b) return TRUE;
+
+  if (!a || !b)
+  {
+    DEBUG ("one is NULL");
+    return FALSE;
+  }
+
+  if (safe_strcmp(a->namespace, b->namespace) != 0)
+  {
+    DEBUG ("namespaces differ: %s vs %s", a->namespace, b->namespace);
+    return FALSE;
+  }
+
+  if (safe_strcmp(a->mnemonic, b->mnemonic) != 0)
+  {
+    DEBUG ("mnemonics differ: %s vs %s", a->mnemonic, b->mnemonic);
+    return FALSE;
+  }
+
+  if (safe_strcmp(a->fullname, b->fullname) != 0)
+  {
+    DEBUG ("fullnames differ: %s vs %s", a->fullname, b->fullname);
+    return FALSE;
+  }
+
+  if (safe_strcmp(a->exchange_code, b->exchange_code) != 0)
+  {
+    DEBUG ("exchange codes differ: %s vs %s",
+           a->exchange_code, b->exchange_code);
+    return FALSE;
+  }
+
+  if (a->fraction != b->fraction)
+  {
+    DEBUG ("fractions differ: %d vs %d", a->fraction, b->fraction);
+    return FALSE;
+  }
+
+  return TRUE;
+}
+
 
 /********************************************************************
  * gnc_commodity_table_new
@@ -309,7 +397,8 @@ gnc_commodity_equiv(const gnc_commodity * a, const gnc_commodity * b) {
  ********************************************************************/
 
 gnc_commodity_table *
-gnc_commodity_table_new(void) {
+gnc_commodity_table_new(void) 
+{
   gnc_commodity_table * retval = g_new0(gnc_commodity_table, 1);
   retval->table = g_hash_table_new(&g_str_hash, &g_str_equal);
   return retval;
@@ -364,12 +453,26 @@ gnc_commodity_table_get_size(gnc_commodity_table* tbl)
 
 gnc_commodity *
 gnc_commodity_table_lookup(const gnc_commodity_table * table, 
-                           const char * namespace, const char * mnemonic) {
+                           const char * namespace, const char * mnemonic)
+{
   gnc_commodity_namespace * nsp = NULL;
+  unsigned int i;
+
+  if (!table || !namespace || !mnemonic) return NULL;
 
   nsp = g_hash_table_lookup(table->table, (gpointer)namespace);
 
   if(nsp) {
+    /*
+     * Backward compatability support for currencies that have
+     * recently changed.
+     */
+    for (i = 0; i < GNC_NEW_ISO_CODES; i++) {
+      if (strcmp(mnemonic, gnc_new_iso_codes[i].old_code) == 0) {
+	mnemonic = gnc_new_iso_codes[i].new_code;
+	break;
+      }
+    }
     return g_hash_table_lookup(nsp->table, (gpointer)mnemonic);
   }
   else {
@@ -377,6 +480,38 @@ gnc_commodity_table_lookup(const gnc_commodity_table * table,
   }
 }
 
+/********************************************************************
+ * gnc_commodity_table_lookup
+ * locate a commodity by unique name.
+ ********************************************************************/
+
+gnc_commodity *
+gnc_commodity_table_lookup_unique(const gnc_commodity_table *table,
+                                  const char * unique_name)
+{
+  char *namespace;
+  char *mnemonic;
+  gnc_commodity *commodity;
+
+  if (!table || !unique_name) return NULL;
+
+  namespace = g_strdup (unique_name);
+  mnemonic = strstr (namespace, "::");
+  if (!mnemonic)
+  {
+    g_free (namespace);
+    return NULL;
+  }
+
+  *mnemonic = '\0';
+  mnemonic += 2;
+
+  commodity = gnc_commodity_table_lookup (table, namespace, mnemonic);
+
+  g_free (namespace);
+
+  return commodity;
+}
 
 /********************************************************************
  * gnc_commodity_table_find_full
@@ -417,7 +552,8 @@ gnc_commodity_table_find_full(const gnc_commodity_table * table,
 
 gnc_commodity *
 gnc_commodity_table_insert(gnc_commodity_table * table, 
-                           gnc_commodity * comm) {
+                           gnc_commodity * comm) 
+{
   gnc_commodity_namespace * nsp = NULL;
   gnc_commodity *c;
 
@@ -488,7 +624,8 @@ gnc_commodity_table_remove(gnc_commodity_table * table,
 
 int
 gnc_commodity_table_has_namespace(const gnc_commodity_table * table,
-                                  const char * namespace) {
+                                  const char * namespace) 
+{
   gnc_commodity_namespace * nsp = NULL;
   
   if(!table || !namespace) { return 0; }
@@ -503,26 +640,30 @@ gnc_commodity_table_has_namespace(const gnc_commodity_table * table,
 }
 
 static void 
-hash_keys_helper(gpointer key, gpointer value, gpointer data) {
+hash_keys_helper(gpointer key, gpointer value, gpointer data) 
+{
   GList ** l = data;
   *l = g_list_prepend(*l, key);
 }
 
 static GList *
-g_hash_table_keys(GHashTable * table) {
+g_hash_table_keys(GHashTable * table) 
+{
   GList * l = NULL;
   g_hash_table_foreach(table, &hash_keys_helper, (gpointer) &l);
   return l;
 }
 
 static void 
-hash_values_helper(gpointer key, gpointer value, gpointer data) {
+hash_values_helper(gpointer key, gpointer value, gpointer data) 
+{
   GList ** l = data;
   *l = g_list_prepend(*l, value);
 }
 
 static GList *
-g_hash_table_values(GHashTable * table) {
+g_hash_table_values(GHashTable * table) 
+{
   GList * l = NULL;
   g_hash_table_foreach(table, &hash_values_helper, (gpointer) &l);
   return l;
@@ -534,7 +675,8 @@ g_hash_table_values(GHashTable * table) {
  ********************************************************************/
 
 GList * 
-gnc_commodity_table_get_namespaces(const gnc_commodity_table * table) {
+gnc_commodity_table_get_namespaces(const gnc_commodity_table * table) 
+{
   if (!table)
     return NULL;
 
@@ -549,7 +691,8 @@ gnc_commodity_table_get_namespaces(const gnc_commodity_table * table) {
 
 GList * 
 gnc_commodity_table_get_commodities(const gnc_commodity_table * table,
-                                    const char * namespace) {
+                                    const char * namespace) 
+{
   gnc_commodity_namespace * ns = NULL; 
 
   if(table) { 
@@ -571,7 +714,8 @@ gnc_commodity_table_get_commodities(const gnc_commodity_table * table,
 
 void 
 gnc_commodity_table_add_namespace(gnc_commodity_table * table,
-                                  const char * namespace) {
+                                  const char * namespace) 
+{
   gnc_commodity_namespace * ns = NULL; 
   
   if(table) { 
@@ -594,7 +738,8 @@ gnc_commodity_table_add_namespace(gnc_commodity_table * table,
  ********************************************************************/
 
 static int
-ns_helper(gpointer key, gpointer value, gpointer user_data) {
+ns_helper(gpointer key, gpointer value, gpointer user_data) 
+{
   gnc_commodity * c = value;
   gnc_commodity_destroy(c);
   g_free(key);
@@ -603,7 +748,8 @@ ns_helper(gpointer key, gpointer value, gpointer user_data) {
 
 void 
 gnc_commodity_table_delete_namespace(gnc_commodity_table * table,
-                                     const char * namespace) {
+                                     const char * namespace) 
+{
   gpointer orig_key;
   gnc_commodity_namespace * value;
 
@@ -684,6 +830,8 @@ gnc_commodity_table_foreach_commodity (gnc_commodity_table * tbl,
 {
   IterData iter_data;
 
+  if (!tbl || !f) return FALSE;
+
   iter_data.ok = TRUE;
   iter_data.func = f;
   iter_data.user_data = user_data;
@@ -699,7 +847,8 @@ gnc_commodity_table_foreach_commodity (gnc_commodity_table * tbl,
  ********************************************************************/
 
 static int
-ct_helper(gpointer key, gpointer value, gpointer data) {
+ct_helper(gpointer key, gpointer value, gpointer data) 
+{
   gnc_commodity_namespace * ns = value;
   g_hash_table_foreach_remove(ns->table, ns_helper, NULL);
   g_hash_table_destroy(ns->table);
@@ -710,7 +859,8 @@ ct_helper(gpointer key, gpointer value, gpointer data) {
 }
 
 void
-gnc_commodity_table_destroy(gnc_commodity_table * t) {
+gnc_commodity_table_destroy(gnc_commodity_table * t) 
+{
   if (!t) return;
   
   g_hash_table_foreach_remove(t->table, ct_helper, t);
@@ -718,4 +868,58 @@ gnc_commodity_table_destroy(gnc_commodity_table * t) {
   g_free(t);
 }
 
+static gboolean 
+table_equal_helper (gnc_commodity *cm_1, gpointer user_data)
+{
+  gnc_commodity_table *t_2 = user_data;
+  gnc_commodity *cm_2;
+
+  cm_2 = gnc_commodity_table_lookup (t_2,
+                                     gnc_commodity_get_namespace (cm_1),
+                                     gnc_commodity_get_mnemonic (cm_1));
+
+  if (!cm_2)
+  {
+    PWARN ("one has commodity %s, the other does not",
+           gnc_commodity_get_unique_name (cm_1));
+    return FALSE;
+  }
+
+  return gnc_commodity_equal (cm_1, cm_2);
+}
+
+gboolean
+gnc_commodity_table_equal(gnc_commodity_table *t_1,
+                          gnc_commodity_table *t_2)
+{
+  gboolean ok;
+
+  if (t_1 == t_2) return TRUE;
+  if (!t_1 || !t_2) return FALSE;
+
+  ok = gnc_commodity_table_foreach_commodity (t_1, table_equal_helper, t_2);
+  if (!ok)
+    return FALSE;
+
+  return gnc_commodity_table_foreach_commodity (t_2, table_equal_helper, t_1);
+}
+
+/********************************************************************
+ * gnc_commodity_table_add_default_data
+ ********************************************************************/
+
+gboolean
+gnc_commodity_table_add_default_data(gnc_commodity_table *table)
+{
+
+  #include "iso-4217-currencies.c"
+
+  gnc_commodity_table_add_namespace(table, GNC_COMMODITY_NS_AMEX);
+  gnc_commodity_table_add_namespace(table, GNC_COMMODITY_NS_NYSE);
+  gnc_commodity_table_add_namespace(table, GNC_COMMODITY_NS_NASDAQ);
+  gnc_commodity_table_add_namespace(table, GNC_COMMODITY_NS_EUREX);
+  gnc_commodity_table_add_namespace(table, GNC_COMMODITY_NS_MUTUAL);
+  return TRUE;
+}
+
 /* ========================= END OF FILE ============================== */
