@@ -26,14 +26,13 @@
 #include <glib.h>
 #include <string.h>
 
-#include "Backend.h"
-#include "GNCId.h"
+#include "BackendP.h"
 #include "GNCIdP.h"
+#include "gnc-book-p.h"
 #include "gnc-engine.h"
 #include "gnc-engine-util.h"
-#include "gnc-event.h"
+#include "gnc-book-p.h"
 #include "gnc-event-p.h"
-#include "gnc-pricedb.h"
 #include "gnc-pricedb-p.h"
 #include "guid.h"
 
@@ -49,19 +48,29 @@ static gboolean remove_price(GNCPriceDB *db, GNCPrice *p, gboolean cleanup);
 
 /* allocation */
 GNCPrice *
-gnc_price_create(void)
+gnc_price_create (GNCBook *book)
 {
-  GNCPrice *p = g_new0(GNCPrice, 1);
-  ENTER(" ");
+  GNCPrice *p;
+
+  g_return_val_if_fail (book, NULL);
+
+  p = g_new0(GNCPrice, 1);
+
   p->refcount = 1;
   p->editlevel = 0;
   p->not_saved = FALSE;
   p->do_free = FALSE;
   p->version = 0;
   p->version_check = 0;
-  xaccGUIDNew (&p->guid);
-  xaccStoreEntity(p, &p->guid, GNC_ID_PRICE); 
+  p->value = gnc_numeric_zero();
+
+  p->book = book;
+  p->entity_table = gnc_book_get_entity_table (book);
+
+  xaccGUIDNew (&p->guid, book);
+  xaccStoreEntity (p->entity_table, p, &p->guid, GNC_ID_PRICE); 
   gnc_engine_generate_event (&p->guid, GNC_EVENT_CREATE);
+
   return p;
 }
 
@@ -70,7 +79,7 @@ gnc_price_destroy (GNCPrice *p)
 {
   ENTER(" ");
   gnc_engine_generate_event (&p->guid, GNC_EVENT_DESTROY);
-  xaccRemoveEntity(&p->guid);
+  xaccRemoveEntity(p->entity_table, &p->guid);
 
   if(p->type) g_cache_remove(gnc_engine_get_string_cache(), p->type);
   if(p->source) g_cache_remove(gnc_engine_get_string_cache(), p->source);
@@ -93,7 +102,7 @@ gnc_price_unref(GNCPrice *p)
   ENTER("pr=%p refcount=%d", p, p->refcount);
   if(p->refcount == 0) {
     PERR("refcount == 0 !!!!");
-    assert(p->refcount != 0);
+    return;
   }
 
   p->refcount--;
@@ -109,14 +118,18 @@ gnc_price_unref(GNCPrice *p)
 /* ==================================================================== */
 
 GNCPrice *
-gnc_price_clone(GNCPrice* p)
+gnc_price_clone (GNCPrice* p, GNCBook *book)
 {
   /* the clone doesn't belong to a PriceDB */
   GNCPrice *new_p;
-  
+
   ENTER ("pr=%p", p);
+
+  g_return_val_if_fail (book, NULL);
+
   if(!p) return NULL;
-  new_p = gnc_price_create();
+
+  new_p = gnc_price_create(book);
   if(!new_p) return NULL;
 
   new_p->version = p->version;
@@ -155,8 +168,8 @@ gnc_price_begin_edit (GNCPrice *p)
   if (p->db) {
     Backend *be;
     be = xaccPriceDBGetBackend (p->db);
-    if (be && be->price_begin_edit) {
-       (be->price_begin_edit) (be, p);
+    if (be && be->begin) {
+       (be->begin) (be, GNC_ID_PRICE, p);
     }
     p->not_saved = FALSE;
   } else {
@@ -186,7 +199,7 @@ gnc_price_commit_edit (GNCPrice *p)
   if (p->db) {
     Backend *be;
     be = xaccPriceDBGetBackend (p->db);
-    if (be && be->price_commit_edit) {
+    if (be && be->commit) {
       GNCBackendError errcode;
 
       /* clear errors */
@@ -196,12 +209,12 @@ gnc_price_commit_edit (GNCPrice *p)
 
       /* if we haven't been able to call begin edit before, call it now */
       if (TRUE == p->not_saved) {
-        if (be->price_begin_edit) {
-          (be->price_begin_edit) (be, p);
+        if (be->begin) {
+          (be->begin) (be, GNC_ID_PRICE, p);
         }
       }
 
-      (be->price_commit_edit) (be, p);
+      (be->commit) (be, GNC_ID_PRICE, p);
       errcode = xaccBackendGetError (be);
       if (ERR_BACKEND_NO_ERR != errcode) 
       {
@@ -227,10 +240,10 @@ void
 gnc_price_set_guid (GNCPrice *p, const GUID *guid)
 {
    if (!p || !guid) return;
-   xaccRemoveEntity (&p->guid);
+   xaccRemoveEntity (p->entity_table, &p->guid);
    p->guid = *guid;
    if(p->db) p->db->dirty = TRUE;
-   xaccStoreEntity(p, &p->guid, GNC_ID_PRICE);
+   xaccStoreEntity(p->entity_table, p, &p->guid, GNC_ID_PRICE);
 }
 
 void
@@ -243,12 +256,14 @@ gnc_price_set_commodity(GNCPrice *p, gnc_commodity *c)
     /* Changing the commodity requires the hash table 
      * position to be modified. The easiest way of doing 
      * this is to remove and reinsert. */
-    remove_price (p->db, p, FALSE);
+    gnc_price_ref (p);
+    remove_price (p->db, p, TRUE);
     gnc_price_begin_edit (p);
     p->commodity = c;
     if(p->db) p->db->dirty = TRUE;
     gnc_price_commit_edit (p);
     add_price (p->db, p);
+    gnc_price_unref (p);
   }
 }
 
@@ -257,17 +272,20 @@ void
 gnc_price_set_currency(GNCPrice *p, gnc_commodity *c)
 {
   if(!p) return;
+
   if(!gnc_commodity_equiv(p->currency, c)) 
   {
     /* Changing the currency requires the hash table 
      * position to be modified. The easiest way of doing 
      * this is to remove and reinsert. */
-    remove_price (p->db, p, FALSE);
+    gnc_price_ref (p);
+    remove_price (p->db, p, TRUE);
     gnc_price_begin_edit (p);
     p->currency = c;
     if(p->db) p->db->dirty = TRUE;
     gnc_price_commit_edit (p);
     add_price (p->db, p);
+    gnc_price_unref (p);
   }
 }
 
@@ -275,17 +293,19 @@ void
 gnc_price_set_time(GNCPrice *p, Timespec t)
 {
   if(!p) return;
-  if(!timespec_equal(&(p->time), &t)) 
+  if(!timespec_equal(&(p->tmspec), &t)) 
   {
-    /* Changing the datesatamp requires the hash table 
+    /* Changing the datestamp requires the hash table 
      * position to be modified. The easiest way of doing 
      * this is to remove and reinsert. */
+    gnc_price_ref (p);
     remove_price (p->db, p, FALSE);
     gnc_price_begin_edit (p);
-    p->time = t;
+    p->tmspec = t;
     if(p->db) p->db->dirty = TRUE;
     gnc_price_commit_edit (p);
     add_price (p->db, p);
+    gnc_price_unref (p);
   }
 }
 
@@ -353,10 +373,12 @@ gnc_price_set_version(GNCPrice *p, gint32 vers)
 /* getters */
 
 GNCPrice *
-gnc_price_lookup (const GUID *guid)
+gnc_price_lookup (const GUID *guid, GNCBook *book)
 {
   if (!guid) return NULL;
-  return xaccLookupEntity (guid, GNC_ID_PRICE);
+  g_return_val_if_fail (book, NULL);
+  return xaccLookupEntity (gnc_book_get_entity_table (book),
+                           guid, GNC_ID_PRICE);
 }
 
 const GUID *
@@ -366,6 +388,20 @@ gnc_price_get_guid (GNCPrice *p)
   return &p->guid;
 }
 
+const GUID
+gnc_price_return_guid (GNCPrice *p)
+{
+  if (!p) return *xaccGUIDNULL();
+  return p->guid;
+}
+
+GNCBook *
+gnc_price_get_book (GNCPrice *p)
+{
+  if (!p) return NULL;
+  return p->book;
+}
+
 gnc_commodity *
 gnc_price_get_commodity(GNCPrice *p)
 {
@@ -382,7 +418,7 @@ gnc_price_get_time(GNCPrice *p)
     result.tv_nsec = 0;
     return result;
   }
-  return p->time;
+  return p->tmspec;
 }
 
 const char *
@@ -423,6 +459,44 @@ gnc_price_get_version(GNCPrice *p)
   return (p->version);
 }
 
+gboolean
+gnc_price_equal (GNCPrice *p1, GNCPrice *p2)
+{
+  Timespec ts1;
+  Timespec ts2;
+
+  if (p1 == p2) return TRUE;
+  if (!p1 || !p2) return FALSE;
+
+  if (!gnc_commodity_equiv (gnc_price_get_commodity (p1),
+                            gnc_price_get_commodity (p2)))
+    return FALSE;
+
+  if (!gnc_commodity_equiv (gnc_price_get_currency (p1),
+                            gnc_price_get_currency (p2)))
+    return FALSE;
+
+  ts1 = gnc_price_get_time (p1);
+  ts2 = gnc_price_get_time (p2);
+
+  if (!timespec_equal (&ts1, &ts2))
+    return FALSE;
+
+  if (safe_strcmp (gnc_price_get_source (p1),
+                   gnc_price_get_source (p2)) != 0)
+    return FALSE;
+
+  if (safe_strcmp (gnc_price_get_type (p1),
+                   gnc_price_get_type (p2)) != 0)
+    return FALSE;
+
+  if (!gnc_numeric_eq (gnc_price_get_value (p1),
+                       gnc_price_get_value (p2)))
+    return FALSE;
+
+  return TRUE;
+}
+
 /* ==================================================================== */
 /* price list manipulation functions */
 
@@ -493,6 +567,32 @@ gnc_price_list_destroy(GList *prices)
   g_list_free(prices);
 }
 
+gboolean
+gnc_price_list_equal(GList *prices1, GList *prices2)
+{
+  GList *n1, *n2;
+
+  if (prices1 == prices2) return TRUE;
+
+  if (g_list_length (prices1) < g_list_length (prices2))
+  {
+    PWARN ("prices2 has extra prices");
+    return FALSE;
+  }
+
+  if (g_list_length (prices1) > g_list_length (prices2))
+  {
+    PWARN ("prices1 has extra prices");
+    return FALSE;
+  }
+
+  for (n1 = prices1, n2 = prices2; n1 ; n1 = n1->next, n2 = n2->next)
+    if (!gnc_price_equal (n1->data, n2->data))
+      return FALSE;
+
+  return TRUE;
+}
+
 /* ==================================================================== */
 /* GNCPriceDB functions
 
@@ -505,12 +605,37 @@ gnc_price_list_destroy(GList *prices)
 
  */
 
-GNCPriceDB *
-gnc_pricedb_create(void)
+static guint
+commodity_hash (gconstpointer key)
 {
-  GNCPriceDB * result = g_new0(GNCPriceDB, 1);
-  result->backend = NULL;
-  result->commodity_hash = g_hash_table_new(g_direct_hash, g_direct_equal);
+  gnc_commodity * com = (gnc_commodity *) key;
+
+  g_return_val_if_fail (key, 0);
+
+  return g_str_hash (gnc_commodity_get_unique_name (com));
+}
+
+static gint
+commodity_equal (gconstpointer a, gconstpointer b)
+{
+  gnc_commodity * ca = (gnc_commodity *) a;
+  gnc_commodity * cb = (gnc_commodity *) b;
+
+  g_return_val_if_fail (a && b, FALSE);
+
+  return gnc_commodity_equiv (ca, cb);
+}
+
+GNCPriceDB *
+gnc_pricedb_create(GNCBook * book)
+{
+  GNCPriceDB * result;
+
+  g_return_val_if_fail (book, NULL);
+
+  result = g_new0(GNCPriceDB, 1);
+  result->book = book;
+  result->commodity_hash = g_hash_table_new(commodity_hash, commodity_equal);
   g_return_val_if_fail (result->commodity_hash, NULL);
   return result;
 }
@@ -555,7 +680,7 @@ gnc_pricedb_destroy(GNCPriceDB *db)
                         NULL);
   g_hash_table_destroy (db->commodity_hash);
   db->commodity_hash = NULL;
-  db->backend = NULL;
+  db->book = NULL;
   g_free(db);
 }
 
@@ -575,6 +700,96 @@ gnc_pricedb_mark_clean(GNCPriceDB *p)
   p->dirty = FALSE;
 }
 
+/* ==================================================================== */
+
+static gboolean
+num_prices_helper (GNCPrice *p, gpointer user_data)
+{
+  guint *count = user_data;
+
+  *count += 1;
+
+  return TRUE;
+}
+
+guint
+gnc_pricedb_get_num_prices(GNCPriceDB *db)
+{
+  guint count;
+
+  if (!db) return 0;
+
+  count = 0;
+
+  gnc_pricedb_foreach_price(db, num_prices_helper, &count, FALSE);
+
+  return count;
+}
+
+/* ==================================================================== */
+
+typedef struct
+{
+  gboolean equal;
+  GNCPriceDB *db2;
+  gnc_commodity *commodity;
+} GNCPriceDBEqualData;
+
+static void
+pricedb_equal_foreach_pricelist(gpointer key, gpointer val, gpointer user_data)
+{
+  GNCPriceDBEqualData *equal_data = user_data;
+  gnc_commodity *currency = key;
+  GList *price_list1 = val;
+  GList *price_list2;
+
+  price_list2 = gnc_pricedb_get_prices (equal_data->db2,
+                                        equal_data->commodity,
+                                        currency);
+
+  if (!gnc_price_list_equal (price_list1, price_list2))
+    equal_data->equal = FALSE;
+
+  gnc_price_list_destroy (price_list2);
+}
+
+static void
+pricedb_equal_foreach_currencies_hash (gpointer key, gpointer val,
+                                       gpointer user_data)
+{
+  GHashTable *currencies_hash = val;
+  GNCPriceDBEqualData *equal_data = user_data;
+
+  equal_data->commodity = key;
+
+  g_hash_table_foreach (currencies_hash,
+                        pricedb_equal_foreach_pricelist,
+                        equal_data);
+}
+
+gboolean
+gnc_pricedb_equal (GNCPriceDB *db1, GNCPriceDB *db2)
+{
+  GNCPriceDBEqualData equal_data;
+
+  if (db1 == db2) return TRUE;
+
+  if (!db1 || !db2)
+  {
+    PWARN ("one is NULL");
+    return FALSE;
+  }
+
+  equal_data.equal = TRUE;
+  equal_data.db2 = db2;
+
+  g_hash_table_foreach (db1->commodity_hash,
+                        pricedb_equal_foreach_currencies_hash,
+                        &equal_data);
+
+  return equal_data.equal;
+}
+
 /* ==================================================================== */
 /* The add_price() function is a utility that only manages the 
  * dual hash table instertion */
@@ -582,7 +797,7 @@ gnc_pricedb_mark_clean(GNCPriceDB *p)
 static gboolean
 add_price(GNCPriceDB *db, GNCPrice *p)
 {
-  /* this function will use p, adding a ref, so treat p as read-only
+  /* This function will use p, adding a ref, so treat p as read-only
      if this function succeeds. */
   GList *price_list;
   gnc_commodity *commodity;
@@ -590,7 +805,17 @@ add_price(GNCPriceDB *db, GNCPrice *p)
   GHashTable *currency_hash;
 
   if(!db || !p) return FALSE;
-  ENTER ("db=%p, pr=%p not-saved=%d do-free=%d", db, p, p->not_saved, p->do_free);
+  ENTER ("db=%p, pr=%p not-saved=%d do-free=%d",
+         db, p, p->not_saved, p->do_free);
+
+  /* initialize the book pointer for teh first time, if needed */
+  if (NULL == db->book) db->book = p->book;
+  if (db->book != p->book)
+  {
+     PERR ("attempted to mix up prices across different books");
+     return FALSE;
+  }
+
   commodity = gnc_price_get_commodity(p);
   if(!commodity) {
     PWARN("no commodity");
@@ -605,7 +830,7 @@ add_price(GNCPriceDB *db, GNCPrice *p)
 
   currency_hash = g_hash_table_lookup(db->commodity_hash, commodity);
   if(!currency_hash) {
-    currency_hash = g_hash_table_new(g_direct_hash, g_direct_equal);
+    currency_hash = g_hash_table_new(commodity_hash, commodity_equal);
     g_hash_table_insert(db->commodity_hash, commodity, currency_hash);
   }
 
@@ -615,18 +840,23 @@ add_price(GNCPriceDB *db, GNCPrice *p)
   g_hash_table_insert(currency_hash, currency, price_list);
   p->db = db;
 
-  LEAVE ("db=%p, pr=%p not-saved=%d do-free=%d", db, p, p->not_saved, p->do_free);
+  LEAVE ("db=%p, pr=%p not-saved=%d do-free=%d commodity=%s/%s currency_hash=%p",
+         db, p, p->not_saved, p->do_free,
+	 gnc_commodity_get_namespace(p->commodity),
+	 gnc_commodity_get_mnemonic(p->commodity),
+	 currency_hash);
   return TRUE;
 }
 
-/* the gnc_pricedb_add_price() function will use p, adding a ref, so treat p as read-only
-   if this function succeeds. (Huh ???) */
+/* the gnc_pricedb_add_price() function will use p, adding a ref, so
+   treat p as read-only if this function succeeds. (Huh ???) */
 gboolean
 gnc_pricedb_add_price(GNCPriceDB *db, GNCPrice *p)
 {
-
   if(!db || !p) return FALSE;
-  ENTER ("db=%p, pr=%p not-saved=%d do-free=%d", db, p, p->not_saved, p->do_free);
+
+  ENTER ("db=%p, pr=%p not-saved=%d do-free=%d",
+         db, p, p->not_saved, p->do_free);
 
   if (FALSE == add_price(db, p)) return FALSE;
 
@@ -636,7 +866,10 @@ gnc_pricedb_add_price(GNCPriceDB *db, GNCPrice *p)
     db->dirty = TRUE;
     gnc_price_commit_edit(p);
   }
-  LEAVE ("db=%p, pr=%p not-saved=%d do-free=%d", db, p, p->not_saved, p->do_free);
+
+  LEAVE ("db=%p, pr=%p not-saved=%d do-free=%d",
+         db, p, p->not_saved, p->do_free);
+
   return TRUE;
 }
 
@@ -653,7 +886,9 @@ remove_price(GNCPriceDB *db, GNCPrice *p, gboolean cleanup)
   GHashTable *currency_hash;
 
   if(!db || !p) return FALSE;
-  ENTER ("db=%p, pr=%p not-saved=%d do-free=%d", db, p, p->not_saved, p->do_free);
+  ENTER ("db=%p, pr=%p not-saved=%d do-free=%d",
+         db, p, p->not_saved, p->do_free);
+
   commodity = gnc_price_get_commodity(p);
   if(!commodity) return FALSE;
   currency = gnc_price_get_currency(p);
@@ -670,15 +905,17 @@ remove_price(GNCPriceDB *db, GNCPrice *p, gboolean cleanup)
     return FALSE;
   }
 
-  /* if the price list is empty, then remove this currency from the commodity hash */
+  /* if the price list is empty, then remove this currency from the
+     commodity hash */
   if(price_list) {
     g_hash_table_insert(currency_hash, currency, price_list);
   } else {
     g_hash_table_remove(currency_hash, currency);
 
     if (cleanup) {
-      /* chances are good that this commodity had only one currency ... 
-       * if there are no currencies, we may as well destroy the commodity  too. */
+      /* chances are good that this commodity had only one currency.
+       * If there are no currencies, we may as well destroy the
+       * commodity too. */
       guint num_currencies = g_hash_table_size (currency_hash);
       if (0 == num_currencies) {
         g_hash_table_remove (db->commodity_hash, commodity);
@@ -697,7 +934,8 @@ gnc_pricedb_remove_price(GNCPriceDB *db, GNCPrice *p)
 {
   gboolean rc;
   if(!db || !p) return FALSE;
-  ENTER ("db=%p, pr=%p not-saved=%d do-free=%d", db, p, p->not_saved, p->do_free);
+  ENTER ("db=%p, pr=%p not-saved=%d do-free=%d",
+         db, p, p->not_saved, p->do_free);
 
   gnc_price_ref(p);
   rc = remove_price (db, p, TRUE);
@@ -729,14 +967,14 @@ gnc_pricedb_lookup_latest(GNCPriceDB *db,
   ENTER ("db=%p commodity=%p currency=%p", db, commodity, currency);
   if(!db || !commodity || !currency) return NULL;
 
-  if (db->backend && db->backend->price_lookup)
+  if (db->book && db->book->backend && db->book->backend->price_lookup)
   {
      GNCPriceLookup pl;
      pl.type = LOOKUP_LATEST;
      pl.prdb = db;
      pl.commodity = commodity;
      pl.currency = currency;
-     (db->backend->price_lookup) (db->backend, &pl);
+     (db->book->backend->price_lookup) (db->book->backend, &pl);
   }
 
   currency_hash = g_hash_table_lookup(db->commodity_hash, commodity);
@@ -745,14 +983,72 @@ gnc_pricedb_lookup_latest(GNCPriceDB *db,
   price_list = g_hash_table_lookup(currency_hash, currency);
   if(!price_list) return NULL;
 
-  /* This works magically because prices are inserted in date-sorted order,
-   * and the latest date always comes first. So return the first in the list.  */
+  /* This works magically because prices are inserted in date-sorted
+   * order, and the latest date always comes first. So return the
+   * first in the list.  */
   result = price_list->data;
   gnc_price_ref(result);
   LEAVE(" ");
   return result;
 }
 
+
+static void
+lookup_latest(gpointer key, gpointer val, gpointer user_data)
+{
+  //gnc_commodity *currency = (gnc_commodity *)key;
+  GList *price_list = (GList *)val;
+  GList **return_list = (GList **)user_data;
+
+  if(!price_list) return;
+
+  /* the latest price is the first in list */
+  gnc_price_list_insert(return_list, price_list->data);
+}
+
+GList *
+gnc_pricedb_lookup_latest_any_currency(GNCPriceDB *db,
+                                       gnc_commodity *commodity)
+{
+  GList *result;
+  GHashTable *currency_hash;
+
+  result = NULL;
+
+  ENTER ("db=%p commodity=%p", db, commodity);
+  if(!db || !commodity) return NULL;
+
+  if (db->book && db->book->backend && db->book->backend->price_lookup)
+  {
+     GNCPriceLookup pl;
+     pl.type = LOOKUP_LATEST;
+     pl.prdb = db;
+     pl.commodity = commodity;
+     pl.currency = NULL;  /* can the backend handle this??? */
+     (db->book->backend->price_lookup) (db->book->backend, &pl);
+  }
+
+  currency_hash = g_hash_table_lookup(db->commodity_hash, commodity);
+  if(!currency_hash) return NULL;
+
+  g_hash_table_foreach(currency_hash, lookup_latest, &result);
+
+  if(!result) return NULL;
+
+  result = g_list_sort(result, compare_prices_by_date);
+
+  LEAVE(" ");
+  return result;
+}
+
+
+static void 
+hash_values_helper(gpointer key, gpointer value, gpointer data) 
+{
+  GList ** l = data;
+  *l = g_list_concat(*l, g_list_copy (value));
+}
+
 GList *
 gnc_pricedb_get_prices(GNCPriceDB *db,
                        gnc_commodity *commodity,
@@ -764,25 +1060,29 @@ gnc_pricedb_get_prices(GNCPriceDB *db,
   GHashTable *currency_hash;
 
   ENTER ("db=%p commodity=%p currency=%p", db, commodity, currency);
-  if(!db || !commodity || !currency) return NULL;
+  if(!db || !commodity) return NULL;
 
-  if (db->backend && db->backend->price_lookup)
+  if (db->book && db->book->backend && db->book->backend->price_lookup)
   {
      GNCPriceLookup pl;
      pl.type = LOOKUP_ALL;
      pl.prdb = db;
      pl.commodity = commodity;
      pl.currency = currency;
-     (db->backend->price_lookup) (db->backend, &pl);
+     (db->book->backend->price_lookup) (db->book->backend, &pl);
   }
 
   currency_hash = g_hash_table_lookup(db->commodity_hash, commodity);
   if(!currency_hash) return NULL;
 
-  price_list = g_hash_table_lookup(currency_hash, currency);
-  if(!price_list) return NULL;
-
-  result = g_list_copy (price_list);
+  if (currency) {
+    price_list = g_hash_table_lookup(currency_hash, currency);
+    if(!price_list) return NULL;
+    result = g_list_copy (price_list);
+  } else {
+    result = NULL;
+    g_hash_table_foreach(currency_hash, hash_values_helper, (gpointer)&result);
+  }
   for (node = result; node; node = node->next)
     gnc_price_ref (node->data);
 
@@ -790,6 +1090,119 @@ gnc_pricedb_get_prices(GNCPriceDB *db,
   return result;
 }
 
+
+GList *
+gnc_pricedb_lookup_day(GNCPriceDB *db,
+		       gnc_commodity *c,
+		       gnc_commodity *currency,
+		       Timespec t)
+{
+  GList *price_list;
+  GList *result = NULL;
+  GList *item = NULL;
+  GHashTable *currency_hash;
+
+  ENTER ("db=%p commodity=%p currency=%p", db, c, currency);
+  if(!db || !c || !currency) return NULL;
+
+  /* Convert to noon local time. */
+  t = timespecCanonicalDayTime(t);
+
+  if (db->book && db->book->backend && db->book->backend->price_lookup)
+  {
+     GNCPriceLookup pl;
+     pl.type = LOOKUP_AT_TIME;
+     pl.prdb = db;
+     pl.commodity = c;
+     pl.currency = currency;
+     pl.date = t;
+     (db->book->backend->price_lookup) (db->book->backend, &pl);
+  }
+
+  currency_hash = g_hash_table_lookup(db->commodity_hash, c);
+  if(!currency_hash) return NULL;
+
+  price_list = g_hash_table_lookup(currency_hash, currency);
+  if(!price_list) return NULL;
+
+  item = price_list;
+  while(item) {
+    GNCPrice *p = item->data;
+    Timespec price_time = timespecCanonicalDayTime(gnc_price_get_time(p));
+    if(timespec_equal(&price_time, &t)) {
+      result = g_list_prepend(result, p);
+      gnc_price_ref(p);
+    }
+    item = item->next;
+  }
+  LEAVE (" ");
+  return result;
+}
+
+
+static void
+lookup_day(gpointer key, gpointer val, gpointer user_data)
+{
+  //gnc_commodity *currency = (gnc_commodity *)key;
+  GList *price_list = (GList *)val;
+  GList *item = NULL;
+  GNCPriceLookupHelper *lookup_helper = (GNCPriceLookupHelper *)user_data;
+  GList **return_list = lookup_helper->return_list;
+  Timespec t = lookup_helper->time;
+
+  item = price_list;
+  while(item) {
+    GNCPrice *p = item->data;
+    Timespec price_time = timespecCanonicalDayTime(gnc_price_get_time(p));
+    if(timespec_equal(&price_time, &t)) {
+      gnc_price_list_insert(return_list, item->data);
+    }
+    item = item->next;
+  }
+}
+
+GList *
+gnc_pricedb_lookup_day_any_currency(GNCPriceDB *db,
+ 		                    gnc_commodity *c,
+		                    Timespec t)
+{
+  GList *result = NULL;
+  GHashTable *currency_hash;
+  GNCPriceLookupHelper lookup_helper;
+
+  ENTER ("db=%p commodity=%p", db, c);
+  if(!db || !c) return NULL;
+
+  /* Convert to noon local time. */
+  t = timespecCanonicalDayTime(t);
+
+  if (db->book && db->book->backend && db->book->backend->price_lookup)
+  {
+     GNCPriceLookup pl;
+     pl.type = LOOKUP_AT_TIME;
+     pl.prdb = db;
+     pl.commodity = c;
+     pl.currency = NULL;  /* can the backend handle this??? */
+     pl.date = t;
+     (db->book->backend->price_lookup) (db->book->backend, &pl);
+  }
+
+  currency_hash = g_hash_table_lookup(db->commodity_hash, c);
+  if(!currency_hash) return NULL;
+
+  lookup_helper.return_list = &result;
+  lookup_helper.time = t;
+  g_hash_table_foreach(currency_hash, lookup_day, &lookup_helper);
+
+  if(!result) return NULL;
+
+  result = g_list_sort(result, compare_prices_by_date);
+
+  LEAVE (" ");
+  return result;
+}
+
+
 GList *
 gnc_pricedb_lookup_at_time(GNCPriceDB *db,
                            gnc_commodity *c,
@@ -804,7 +1217,7 @@ gnc_pricedb_lookup_at_time(GNCPriceDB *db,
   ENTER ("db=%p commodity=%p currency=%p", db, c, currency);
   if(!db || !c || !currency) return NULL;
 
-  if (db->backend && db->backend->price_lookup)
+  if (db->book && db->book->backend && db->book->backend->price_lookup)
   {
      GNCPriceLookup pl;
      pl.type = LOOKUP_AT_TIME;
@@ -812,7 +1225,7 @@ gnc_pricedb_lookup_at_time(GNCPriceDB *db,
      pl.commodity = c;
      pl.currency = currency;
      pl.date = t;
-     (db->backend->price_lookup) (db->backend, &pl);
+     (db->book->backend->price_lookup) (db->book->backend, &pl);
   }
 
   currency_hash = g_hash_table_lookup(db->commodity_hash, c);
@@ -835,6 +1248,65 @@ gnc_pricedb_lookup_at_time(GNCPriceDB *db,
   return result;
 }
 
+static void
+lookup_time(gpointer key, gpointer val, gpointer user_data)
+{
+  //gnc_commodity *currency = (gnc_commodity *)key;
+  GList *price_list = (GList *)val;
+  GList *item = NULL;
+  GNCPriceLookupHelper *lookup_helper = (GNCPriceLookupHelper *)user_data;
+  GList **return_list = lookup_helper->return_list;
+  Timespec t = lookup_helper->time;
+
+  item = price_list;
+  while(item) {
+    GNCPrice *p = item->data;
+    Timespec price_time = gnc_price_get_time(p);
+    if(timespec_equal(&price_time, &t)) {
+      gnc_price_list_insert(return_list, item->data);
+    }
+    item = item->next;
+  }
+}
+
+GList *
+gnc_pricedb_lookup_at_time_any_currency(GNCPriceDB *db,
+ 		                        gnc_commodity *c,
+		                        Timespec t)
+{
+  GList *result = NULL;
+  GHashTable *currency_hash;
+  GNCPriceLookupHelper lookup_helper;
+
+  ENTER ("db=%p commodity=%p", db, c);
+  if(!db || !c) return NULL;
+
+  if (db->book && db->book->backend && db->book->backend->price_lookup)
+  {
+     GNCPriceLookup pl;
+     pl.type = LOOKUP_AT_TIME;
+     pl.prdb = db;
+     pl.commodity = c;
+     pl.currency = NULL;  /* can the backend handle this??? */
+     pl.date = t;
+     (db->book->backend->price_lookup) (db->book->backend, &pl);
+  }
+
+  currency_hash = g_hash_table_lookup(db->commodity_hash, c);
+  if(!currency_hash) return NULL;
+
+  lookup_helper.return_list = &result;
+  lookup_helper.time = t;
+  g_hash_table_foreach(currency_hash, lookup_time, &lookup_helper);
+
+  if(!result) return NULL;
+
+  result = g_list_sort(result, compare_prices_by_date);
+
+  LEAVE (" ");
+  return result;
+}
+
 
 GNCPrice *
 gnc_pricedb_lookup_nearest_in_time(GNCPriceDB *db,
@@ -852,7 +1324,7 @@ gnc_pricedb_lookup_nearest_in_time(GNCPriceDB *db,
   ENTER ("db=%p commodity=%p currency=%p", db, c, currency);
   if(!db || !c || !currency) return NULL;
 
-  if (db->backend && db->backend->price_lookup)
+  if (db->book && db->book->backend && db->book->backend->price_lookup)
   {
      GNCPriceLookup pl;
      pl.type = LOOKUP_NEAREST_IN_TIME;
@@ -860,7 +1332,7 @@ gnc_pricedb_lookup_nearest_in_time(GNCPriceDB *db,
      pl.commodity = c;
      pl.currency = currency;
      pl.date = t;
-     (db->backend->price_lookup) (db->backend, &pl);
+     (db->book->backend->price_lookup) (db->book->backend, &pl);
   }
 
   currency_hash = g_hash_table_lookup(db->commodity_hash, c);
@@ -911,6 +1383,253 @@ gnc_pricedb_lookup_nearest_in_time(GNCPriceDB *db,
   return result;
 }
 
+static void
+lookup_nearest(gpointer key, gpointer val, gpointer user_data)
+{
+  //gnc_commodity *currency = (gnc_commodity *)key;
+  GList *price_list = (GList *)val;
+  GNCPrice *current_price = NULL;
+  GNCPrice *next_price = NULL;
+  GNCPrice *result = NULL;
+  GList *item = NULL;
+  GNCPriceLookupHelper *lookup_helper = (GNCPriceLookupHelper *)user_data;
+  GList **return_list = lookup_helper->return_list;
+  Timespec t = lookup_helper->time;
+
+  item = price_list;
+
+  /* default answer */
+  current_price = item->data;
+
+  /* find the first candidate past the one we want.  Remember that
+     prices are in most-recent-first order. */
+  while (!next_price && item) {
+    GNCPrice *p = item->data;
+    Timespec price_time = gnc_price_get_time(p);
+    if (timespec_cmp(&price_time, &t) <= 0) {
+      next_price = item->data;
+      break;
+    }
+    current_price = item->data;
+    item = item->next;
+  }
+
+  if (current_price) {
+    if (!next_price) {
+      result = current_price;
+    } else {
+      Timespec current_t = gnc_price_get_time(current_price);
+      Timespec next_t = gnc_price_get_time(next_price);
+      Timespec diff_current = timespec_diff(&current_t, &t);
+      Timespec diff_next = timespec_diff(&next_t, &t);
+      Timespec abs_current = timespec_abs(&diff_current);
+      Timespec abs_next = timespec_abs(&diff_next);
+      
+      if (timespec_cmp(&abs_current, &abs_next) <= 0) {
+	result = current_price;
+      } else {
+	result = next_price;
+      }
+    }
+  }
+
+  gnc_price_list_insert(return_list, result);
+}
+
+GList *
+gnc_pricedb_lookup_nearest_in_time_any_currency(GNCPriceDB *db,
+ 		                                gnc_commodity *c,
+		                                Timespec t)
+{
+  GList *result = NULL;
+  GHashTable *currency_hash;
+  GNCPriceLookupHelper lookup_helper;
+
+  ENTER ("db=%p commodity=%p", db, c);
+  if(!db || !c) return NULL;
+
+  if (db->book && db->book->backend && db->book->backend->price_lookup)
+  {
+     GNCPriceLookup pl;
+     pl.type = LOOKUP_NEAREST_IN_TIME;
+     pl.prdb = db;
+     pl.commodity = c;
+     pl.currency = NULL;  /* can the backend handle this??? */
+     pl.date = t;
+     (db->book->backend->price_lookup) (db->book->backend, &pl);
+  }
+
+  currency_hash = g_hash_table_lookup(db->commodity_hash, c);
+  if(!currency_hash) return NULL;
+
+  lookup_helper.return_list = &result;
+  lookup_helper.time = t;
+  g_hash_table_foreach(currency_hash, lookup_nearest, &lookup_helper);
+
+  if(!result) return NULL;
+
+  result = g_list_sort(result, compare_prices_by_date);
+
+  LEAVE (" ");
+  return result;
+}
+
+
+/*
+ * Convert a balance from one currency to another.
+ */
+gnc_numeric
+gnc_pricedb_convert_balance_latest_price(GNCPriceDB *pdb,
+				         gnc_numeric balance,
+				         gnc_commodity *balance_currency,
+				         gnc_commodity *new_currency)
+{
+  GNCPrice *price, *currency_price;
+  GList *price_list, *list_helper;
+  gnc_numeric currency_price_value;
+  gnc_commodity *intermediate_currency;
+
+  if (gnc_numeric_zero_p (balance) ||
+      gnc_commodity_equiv (balance_currency, new_currency))
+    return balance;
+
+  /* Look for a direct price. */
+  price = gnc_pricedb_lookup_latest (pdb, balance_currency, new_currency);
+  if (price) {
+    balance = gnc_numeric_mul (balance, gnc_price_get_value (price),
+			       gnc_commodity_get_fraction (new_currency),
+			       GNC_RND_ROUND);
+    gnc_price_unref (price);
+    return balance;
+  }
+
+  /*
+   * no direct price found, try if we find a price in another currency
+   * and convert in two stages
+   */
+  price_list = gnc_pricedb_lookup_latest_any_currency(pdb, balance_currency);
+  if (!price_list) {
+    balance =  gnc_numeric_zero ();
+    return balance;
+  }
+
+  list_helper = price_list;
+  currency_price_value = gnc_numeric_zero();
+
+  do {
+    price = (GNCPrice *)(list_helper->data);
+
+    intermediate_currency = gnc_price_get_currency(price);
+    currency_price = gnc_pricedb_lookup_latest(pdb, intermediate_currency,
+					       new_currency);
+    if(currency_price) {
+      currency_price_value = gnc_price_get_value(currency_price);
+      gnc_price_unref(currency_price);
+    } else {
+      currency_price = gnc_pricedb_lookup_latest(pdb, new_currency,
+						 intermediate_currency);
+      if (currency_price) {
+	/* here we need the reciprocal */
+	currency_price_value = gnc_numeric_div(gnc_numeric_create(1, 1),
+					       gnc_price_get_value(currency_price),
+					       GNC_DENOM_AUTO,
+					       GNC_DENOM_EXACT | GNC_RND_NEVER);
+	gnc_price_unref(currency_price);
+      }
+    }
+
+    list_helper = list_helper->next;
+  } while((list_helper != NULL) &&
+	  (!gnc_numeric_zero_p(currency_price_value)));
+
+  balance = gnc_numeric_mul (balance, currency_price_value,
+			     gnc_commodity_get_fraction (new_currency),
+			     GNC_RND_ROUND);      
+  balance = gnc_numeric_mul (balance, gnc_price_get_value (price),
+			     gnc_commodity_get_fraction (new_currency),
+			     GNC_RND_ROUND);      
+
+  gnc_price_list_destroy(price_list);
+  return balance;
+}
+
+gnc_numeric
+gnc_pricedb_convert_balance_nearest_price(GNCPriceDB *pdb,
+				          gnc_numeric balance,
+				          gnc_commodity *balance_currency,
+				          gnc_commodity *new_currency,
+					  Timespec t)
+{
+  GNCPrice *price, *currency_price;
+  GList *price_list, *list_helper;
+  gnc_numeric currency_price_value;
+  gnc_commodity *intermediate_currency;
+
+  if (gnc_numeric_zero_p (balance) ||
+      gnc_commodity_equiv (balance_currency, new_currency))
+    return balance;
+
+  /* Look for a direct price. */
+  price = gnc_pricedb_lookup_nearest_in_time (pdb, balance_currency, new_currency, t);
+  if (price) {
+    balance = gnc_numeric_mul (balance, gnc_price_get_value (price),
+			       gnc_commodity_get_fraction (new_currency),
+			       GNC_RND_ROUND);
+    gnc_price_unref (price);
+    return balance;
+  }
+
+  /*
+   * no direct price found, try if we find a price in another currency
+   * and convert in two stages
+   */
+  price_list = gnc_pricedb_lookup_nearest_in_time_any_currency(pdb, balance_currency, t);
+  if (!price_list) {
+    balance =  gnc_numeric_zero ();
+    return balance;
+  }
+
+  list_helper = price_list;
+  currency_price_value = gnc_numeric_zero();
+
+  do {
+    price = (GNCPrice *)(list_helper->data);
+
+    intermediate_currency = gnc_price_get_currency(price);
+    currency_price = gnc_pricedb_lookup_nearest_in_time(pdb, intermediate_currency,
+					                new_currency, t);
+    if(currency_price) {
+      currency_price_value = gnc_price_get_value(currency_price);
+      gnc_price_unref(currency_price);
+    } else {
+      currency_price = gnc_pricedb_lookup_nearest_in_time(pdb, new_currency,
+						          intermediate_currency, t);
+      if (currency_price) {
+	/* here we need the reciprocal */
+	currency_price_value = gnc_numeric_div(gnc_numeric_create(1, 1),
+					       gnc_price_get_value(currency_price),
+					       gnc_commodity_get_fraction (new_currency),
+					       GNC_RND_ROUND);
+	gnc_price_unref(currency_price);
+      }
+    }
+
+    list_helper = list_helper->next;
+  } while((list_helper != NULL) &&
+	  (!gnc_numeric_zero_p(currency_price_value)));
+
+  balance = gnc_numeric_mul (balance, currency_price_value,
+			     gnc_commodity_get_fraction (new_currency),
+			     GNC_RND_ROUND);      
+  balance = gnc_numeric_mul (balance, gnc_price_get_value (price),
+			     gnc_commodity_get_fraction (new_currency),
+			     GNC_RND_ROUND);      
+
+  gnc_price_list_destroy(price_list);
+  return balance;
+}
+
+
 /* ==================================================================== */
 /* gnc_pricedb_foreach_price infrastructure
  */
@@ -994,7 +1713,6 @@ stable_price_traversal(GNCPriceDB *db,
                        gpointer user_data)
 {
   GSList *currency_hashes = NULL;
-  GSList *foo = NULL;
   gboolean ok = TRUE;
   GSList *i = NULL;
   
@@ -1059,36 +1777,31 @@ typedef struct {
 } GNCPriceFixupData;
 
 static gboolean
-gnc_price_fixup_legacy_commods(GNCPrice *p, gpointer data)
+add_price_to_list (GNCPrice *p, gpointer data)
 {
-  GNCPriceFixupData *fixup_data = (GNCPriceFixupData *) data;
-  gnc_commodity *price_c;
+  GList **list = data;
 
-  if (!p) return FALSE;
+  *list = g_list_prepend (*list, p);
 
-  price_c = gnc_price_get_commodity(p);
-  if (gnc_commodity_equiv(price_c, fixup_data->old_c)) {
-    gnc_price_set_commodity(p, fixup_data->new_c);
-  }
-  price_c = gnc_price_get_currency(p);
-  if (gnc_commodity_equiv(price_c, fixup_data->old_c)) {
-    gnc_price_set_currency(p, fixup_data->new_c);
-  }
   return TRUE;
 }
 
-
 static void
-remap_currency_hash_keys(gpointer key, gpointer val, gpointer user_data)
+gnc_price_fixup_legacy_commods(gpointer data, gpointer user_data)
 {
-  GHashTable *currencies_hash = (GHashTable *) val;
-  GNCPriceFixupData *fixup_data = (GNCPriceFixupData *) user_data;
-  GList *price_list;
-  
-  price_list = g_hash_table_lookup(currencies_hash, fixup_data->old_c);
-  if(price_list) {
-    g_hash_table_remove(currencies_hash, fixup_data->old_c);
-    g_hash_table_insert(currencies_hash, fixup_data->new_c, price_list);
+  GNCPrice *p = data;
+  GNCPriceFixupData *fixup_data = user_data;
+  gnc_commodity *price_c;
+
+  if (!p) return;
+
+  price_c = gnc_price_get_commodity(p);
+  if (gnc_commodity_equiv(price_c, fixup_data->old_c)) {
+    gnc_price_set_commodity (p, fixup_data->new_c);
+  }
+  price_c = gnc_price_get_currency(p);
+  if (gnc_commodity_equiv(price_c, fixup_data->old_c)) {
+    gnc_price_set_currency (p, fixup_data->new_c);
   }
 }
 
@@ -1097,29 +1810,19 @@ gnc_pricedb_substitute_commodity(GNCPriceDB *db,
                                  gnc_commodity *old_c,
                                  gnc_commodity *new_c)
 {
-  GHashTable *currency_hash;
   GNCPriceFixupData data;
+  GList *prices = NULL;
 
   if(!db || !old_c || !new_c) return;
 
   data.old_c = old_c;
   data.new_c = new_c;
 
-  /* first remap the relevant commodity -> currency hash, if any */
-  currency_hash = g_hash_table_lookup(db->commodity_hash, old_c);
-  if(currency_hash) {
-    g_hash_table_remove(db->commodity_hash, old_c);
-    g_hash_table_insert(db->commodity_hash, new_c, currency_hash);
-  }
+  gnc_pricedb_foreach_price (db, add_price_to_list, &prices, FALSE);
 
-  g_hash_table_foreach(db->commodity_hash, remap_currency_hash_keys, &data);
+  g_list_foreach (prices, gnc_price_fixup_legacy_commods, &data);
 
-  if(!gnc_pricedb_foreach_price(db,
-                                gnc_price_fixup_legacy_commods,
-                                &data,
-                                FALSE)) {
-    PERR("Adjustments to legacy commodity pointers in pricedb failed!");
-  }   
+  g_list_free (prices);
 }
 
 /***************************************************************************/
@@ -1149,7 +1852,7 @@ gnc_price_print(GNCPrice *p, FILE *f, int indent)
   fprintf(f, "%s  <pdb:commodity pointer=%p>\n", istr, commodity);
   str = gnc_commodity_get_namespace(commodity);
   str = str ? str : "(null)";
-  fprintf(f, "%s    <cmdty:ref-space> %s</gnc:cmdty:ref-space>\n", istr, str);
+  fprintf(f, "%s    <cmdty:ref-space>%s</gnc:cmdty:ref-space>\n", istr, str);
   str = gnc_commodity_get_mnemonic(commodity);
   str = str ? str : "(null)";
   fprintf(f, "%s    <cmdty:ref-id>%s</cmdty:ref-id>\n", istr, str);
@@ -1174,12 +1877,6 @@ gnc_price_print(GNCPrice *p, FILE *f, int indent)
   g_free(istr);
 }
 
-static void
-gnc_price_print_stdout(GNCPrice *p, int indent)
-{
-  gnc_price_print(p, stdout, indent);
-}
-
 static gboolean
 print_pricedb_adapter(GNCPrice *p, gpointer user_data)
 {
