@@ -20,8 +20,6 @@
 ;; Boston, MA  02111-1307,  USA       gnu@gnu.org
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-(gnc:support "gnc:main-window.scm")
-
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; account tree options 
 ;; like reports, we have an integer tree id that is the index into a
@@ -58,7 +56,7 @@ the account instead of opening a register.") #f))
       (N_ "Account Tree") (N_ "Account types to display")
       "b" ""
       (list 'bank 'cash 'credit 'asset 'liability 'stock
-            'mutual 'currency 'income 'expense 'equity)
+            'mutual 'currency 'income 'expense 'equity 'payable 'receivable)
       (list (list->vector (list 'bank      (N_ "Bank") ""))
             (list->vector (list 'cash      (N_ "Cash") ""))
             (list->vector (list 'credit    (N_ "Credit") ""))
@@ -69,22 +67,33 @@ the account instead of opening a register.") #f))
             (list->vector (list 'currency  (N_ "Currency") ""))
             (list->vector (list 'income    (N_ "Income") ""))
             (list->vector (list 'expense   (N_ "Expense") ""))
-            (list->vector (list 'equity    (N_ "Equity") "")))))
+            (list->vector (list 'equity    (N_ "Equity") ""))
+	    (list->vector (list 'payable   (N_ "Accounts Payable") ""))
+	    (list->vector (list 'receivable (N_ "Accounts Receivable") "")))))
 
     (add-option
      (gnc:make-list-option
       (N_ "Account Tree") (N_ "Account fields to display")
       "c" ""
       (list 'description 'total)
-      (list (list->vector (list 'type        (N_ "Type") ""))
-            (list->vector (list 'code        (N_ "Code") ""))
-            (list->vector (list 'description (N_ "Description") ""))
-            (list->vector (list 'notes       (N_ "Notes") ""))
-            (list->vector (list 'currency    (N_ "Currency") ""))
-            (list->vector (list 'security    (N_ "Security") ""))
-            (list->vector (list 'balance     (N_ "Balance") ""))
-            (list->vector (list 'total       (N_ "Total") ""))
-            (list->vector (list 'tax-info    (N_ "Tax Info") "")))))
+      (list (list->vector (list 'type           (N_ "Type") ""))
+            (list->vector (list 'code           (N_ "Code") ""))
+            (list->vector (list 'description    (N_ "Description") ""))
+            (list->vector (list 'notes          (N_ "Notes") ""))
+            (list->vector (list 'commodity      (N_ "Commodity") ""))
+            (list->vector (list 'balance        (N_ "Balance") ""))
+            (list->vector (list 'balance_report (N_ "Balance in Report Currency") ""))
+            (list->vector (list 'present        (N_ "Present Balance") ""))
+            (list->vector (list 'present_report (N_ "Present Balance in Report Currency") ""))
+            (list->vector (list 'cleared        (N_ "Cleared Balance") ""))
+            (list->vector (list 'cleared_report (N_ "Cleared Balance in Report Currency") ""))
+            (list->vector (list 'reconciled        (N_ "Reconciled Balance") ""))
+            (list->vector (list 'reconciled_report (N_ "Reconciled Balance in Report Currency") ""))
+            (list->vector (list 'future_min        (N_ "Future Minimum Balance") ""))
+            (list->vector (list 'future_min_report (N_ "Future Minimum Balance in Report Currency") ""))
+            (list->vector (list 'total          (N_ "Total") ""))
+            (list->vector (list 'total_report   (N_ "Total in Report Currency") ""))
+            (list->vector (list 'tax-info       (N_ "Tax Info") "")))))
 
     options))
 
@@ -107,8 +116,9 @@ the account instead of opening a register.") #f))
    (gnc:generate-restore-forms optobj "options")
    (simple-format
     #f "  (hash-set! gnc:*acct-tree-options* ~A options)\n" id)
-   (simple-format
-    #f "  \"gnc-acct-tree:id=~S\")\n\n" id)))
+   "  \""
+   (gnc:html-build-url gnc:url-type-accttree (sprintf #f "%a" id) #f)
+   "\")\n\n"))
 
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -118,8 +128,9 @@ the account instead of opening a register.") #f))
 ;; book close.
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-(define (gnc:main-window-save-state book-url)
-  (let* ((conf-file-name (gnc:html-encode-string book-url))
+(define (gnc:main-window-save-state session)
+  (let* ((book-url (gnc:session-get-url session))
+	 (conf-file-name (gnc:html-encode-string book-url))
 	 (dotgnucash-dir (build-path (getenv "HOME") ".gnucash"))
          (file-dir (build-path dotgnucash-dir "books"))
          (save-file? #f)
@@ -131,8 +142,7 @@ the account instead of opening a register.") #f))
 
     (if (not save-file?) (gnc:warn (_ "Can't save window state")))
 
-    (if (and save-file? conf-file-name
-             (gnc:main-window-can-save? (gnc:get-ui-data)))
+    (if (and save-file? conf-file-name (gnc:mdi-has-apps?))
         (let ((book-path (build-path (getenv "HOME") ".gnucash" "books" 
                                      conf-file-name)))
           (with-output-to-port (open-output-file book-path)
@@ -143,17 +153,17 @@ the account instead of opening a register.") #f))
                      (display 
                       (gnc:report-generate-restore-forms v))))
                #t *gnc:_reports_*)
-              
+
               (hash-fold 
                (lambda (k v p)
                  (display (gnc:acct-tree-generate-restore-forms v k)) #t)
                #t gnc:*acct-tree-options*)
 
               (force-output)))
-          (gnc:main-window-save (gnc:get-ui-data) book-url)))))
+          (gnc:mdi-save (gnc:mdi-get-current) book-url)))))
 
-(define (gnc:main-window-book-close-handler book-url)
-    (gnc:main-window-save-state book-url)
+(define (gnc:main-window-book-close-handler session)
+    (gnc:main-window-save-state session)
 
     (let ((dead-reports '()))
       ;; get a list of the reports we'll be needing to nuke     
@@ -169,7 +179,7 @@ the account instead of opening a register.") #f))
          (hash-remove! *gnc:_reports_* dr))
        dead-reports)))
 
-(define (gnc:main-window-book-open-handler book-url)
+(define (gnc:main-window-book-open-handler session)
   (define (try-load file-suffix)
     (let ((file (build-path (getenv "HOME") ".gnucash" "books" file-suffix)))
       ;; make sure the books directory is there 
@@ -179,14 +189,42 @@ the account instead of opening a register.") #f))
                 (gnc:warn "failure loading " file)
                 #f))
           #f)))
-  (let ((conf-file-name (gnc:html-encode-string book-url))
-        (dead-reports '()))
+  (let* ((book-url (gnc:session-get-url session))
+	 (conf-file-name (gnc:html-encode-string book-url))
+	 (dead-reports '()))
     (if conf-file-name 
         (try-load conf-file-name))
-    (gnc:main-window-restore (gnc:get-ui-data) book-url)))
+    (gnc:mdi-restore (gnc:mdi-get-current) book-url)))
 
+(define (gnc:main-window-ui-startup-handler)
+  (gnc:add-extension
+    (gnc:make-menu-item (N_ "Properties")
+			(N_ "View and edit the properties of this file.")
+			(list "Main" "_File" "_Print")
+			(lambda ()
+			  (let* ((book (gnc:get-current-book))
+				 (slots (gnc:book-get-slots book)))
+
+			    (define (changed_cb)
+			      (gnc:book-kvp-changed book))
+			    
+			    (gnc:kvp-option-dialog gnc:id-book
+						   slots "Book Options"
+						   changed_cb)))))
+  (gnc:add-extension (gnc:make-separator (list "Main" "_File" "_Print")))
+)
+
+(gnc:hook-remove-dangler gnc:*book-opened-hook* 
+                         gnc:main-window-book-open-handler)
 (gnc:hook-add-dangler gnc:*book-opened-hook* 
                       gnc:main-window-book-open-handler)
+
+(gnc:hook-remove-dangler gnc:*book-closed-hook* 
+                         gnc:main-window-book-close-handler)
 (gnc:hook-add-dangler gnc:*book-closed-hook* 
                       gnc:main-window-book-close-handler)
 
+(gnc:hook-remove-dangler gnc:*ui-startup-hook* 
+                         gnc:main-window-ui-startup-handler)
+(gnc:hook-add-dangler gnc:*ui-startup-hook* 
+                      gnc:main-window-ui-startup-handler)
