@@ -20,16 +20,21 @@
 ;;; Boston, MA  02111-1307,  USA       gnu@gnu.org
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-;(use-modules (gnucash process))
+(define-module (gnucash price-quotes))
 
-(gnc:support "price-quotes.scm")
-
-(gnc:depend "process.scm")
-(gnc:depend "gnc-numeric.scm")
-(gnc:depend "utilities.scm")
-(gnc:depend "engine-utilities.scm")
+(export yahoo-get-historical-quotes)
+(export gnc:fq-check-sources)
+(export gnc:book-add-quotes)
+(export gnc:add-quotes-to-book-at-url)
 
+(use-modules (gnucash process))
+(use-modules (www main))
+(use-modules (srfi srfi-1))
+(use-modules (gnucash main) (g-wrapped gw-gnc)) ;; FIXME: delete after we finish modularizing.
+(use-modules (gnucash gnc-module))
+(use-modules (g-wrapped gw-gnome-utils))
 
+(gnc:module-load "gnucash/app-utils" 0)
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;
@@ -206,6 +211,45 @@
                 result)
             #f))))
 
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+
+(define gnc:*finance-quote-check*
+  (gnc:find-file "finance-quote-check"
+                 (gnc:config-var-value-get gnc:*share-path*)))
+
+(define (gnc:fq-check-sources)
+  (let ((program #f))
+
+    (define (start-program)
+      (set! program (gnc:run-sub-process #f
+                                        gnc:*finance-quote-check*
+                                        gnc:*finance-quote-check*)))
+
+    (define (get-sources)
+      (and program
+           (let ((from-child (cadr program))
+                 (results #f))
+	     (catch
+	      #t
+	      (lambda ()
+		(set! results (read from-child))
+;;		(write (list 'results results)) (newline)
+		results)
+	      (lambda (key . args)
+		key)))))
+
+    (define (kill-program)
+      (and program
+           (let ((pid (car program)))
+             (close-input-port (cadr program))
+             (close-output-port (caddr program))
+             (gnc:cleanup-sub-process (car program) 1))))
+
+    (dynamic-wind
+        start-program
+        get-sources
+        kill-program)))
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;
 ;; Finance::Quote based instantaneous quotes -- used by the
@@ -219,8 +263,8 @@
 ;; src/engine/gnc-pricedb.h
 
 (define gnc:*finance-quote-helper*
-  (string-append (gnc:config-var-value-get gnc:*share-dir*)
-                 "/finance-quote-helper"))
+  (gnc:find-file "finance-quote-helper"
+                 (gnc:config-var-value-get gnc:*share-path*)))
 
 (define (gnc:fq-get-quotes requests)
   ;; requests should be a list where each item is of the form
@@ -270,21 +314,28 @@
     (define (get-quotes)
       (and quoter
            (let ((to-child (caddr quoter))
-                 (from-child (cadr quoter)))
+                 (from-child (cadr quoter))
+                 (results #f))
              (map
               (lambda (request)
-                (write (list 'handling-request request))
-                (newline)
-                ;; we need to display the first element (the method, so it
-                ;; won't be quoted) and then write the rest
-                (display #\( to-child)
-                (display (car request) to-child)
-                (display " " to-child)
-                (for-each (lambda (x) (write x to-child)) (cdr request))
-                (display #\) to-child)
-                (newline to-child)
-                (force-output to-child)
-                (read from-child))
+		(catch
+		 #t
+		 (lambda ()
+;;		   (write (list 'handling-request request)) (newline)
+		   ;; we need to display the first element (the method, so it
+		   ;; won't be quoted) and then write the rest
+		   (display #\( to-child)
+		   (display (car request) to-child)
+		   (display " " to-child)
+		   (for-each (lambda (x) (write x to-child)) (cdr request))
+		   (display #\) to-child)
+		   (newline to-child)
+		   (force-output to-child)
+		   (set! results (read from-child))
+;;		   (write (list 'results results)) (newline)
+		   results)
+		 (lambda (key . args)
+		   key)))
              requests))))
 
     (define (kill-quoter)
@@ -305,15 +356,14 @@
     ;; Return a list of accounts for whose commodities we should get
     ;; quotes.
     (define (quotable-currency-account? a)
-      (let ((currency (gnc:account-get-currency a))
-            (security (gnc:account-get-security a)))
-        (and (equal? (gnc:commodity-get-namespace currency) "ISO4217")
-             (equal? (gnc:commodity-get-namespace security) "ISO4217"))))
+      (let ((commodity (gnc:account-get-commodity a)))
+        (equal? (gnc:commodity-get-namespace commodity) "ISO4217")))
 
     (define (quotable-account? a)
       (let ((type (gw:enum-<gnc:AccountType>-val->sym (gnc:account-get-type a)
                                                       #f))
-	    (src (gnc:account-get-price-src a)))
+	    (src (gnc:account-get-price-src a))
+	    (balance (not (gnc:numeric-zero-p (gnc:account-get-balance a)))))
 
         (if (not type) (set! type '()))
         (if (symbol? type) (set! type (list type)))
@@ -325,7 +375,11 @@
 	    a
 	    #f)))
 
-    (filter quotable-account? (gnc:group-get-subaccounts group)))
+    (let ((quotables (filter quotable-account? (gnc:group-get-subaccounts group))))
+      (if (null? quotables)
+	  #f
+	  quotables))
+    )
 
   (define (accounts->fq-call-data account-list)
     ;; Take a list of accounts that should be "quotable" -- i.e. they
@@ -345,43 +399,14 @@
     ;;                     (commodity-4 currency-4 tz-4) ...)
     ;;  ...)
 
-  (define (src->fq-method-sym src)
-    (cond
-     ((string=? "YAHOO" src) 'yahoo)
-     ((string=? "YAHOO_EUROPE" src) 'yahoo-europe)
-     ((string=? "FIDELITY" src) 'fidelity)
-     ((string=? "TRPRICE" src) 'troweprice)
-     ((string=? "VANGUARD" src) 'vanguard)
-     ((string=? "ASX" src) 'asx)
-     ((string=? "TIAACREF" src) 'tiaacref)
-     ((string=? "TRUSTNET" src) 'trustnet)
-     ((string=? "VWD" src) 'vwd)
-     ((string=? "CURRENCY" src) 'currency)
-     (else #f)))
-
-  ;; NOTE: If you modify this, please update finance-quote-helper.in as well.
-  (define (fq-method-sym->str src-sym)
-    (case src-sym
-     ((yahoo) "yahoo")
-     ((yahoo-europe) "yahoo_europe")
-     ((fidelity) "fidelity_direct")
-     ((troweprice) "troweprice_direct")
-     ((vanguard) "vanguard")
-     ((asx) "asx")
-     ((tiaacref) "tiaacref")
-     ((trustnet) "trustnet")
-     ((vwd) "vwd")
-     ((currency) "currency")
-     (else #f)))
-
   (define (account->fq-cmd account)
     ;; Returns (cons fq-method-sym
     ;;               (list commodity currency assumed-timezone-str))
     (let* ((commodity (gnc:account-get-commodity account))
-           (currency (gnc:account-get-currency account))
+           (currency (gnc:default-currency))
            (src (and account (gnc:account-get-price-src account)))
            (tz (gnc:account-get-quote-tz account))
-	   (fq-method-sym (and src (src->fq-method-sym src)))
+	   (fq-method-sym (and src (gnc:price-source-internal2fq src)))
            (mnemonic (and commodity (gnc:commodity-get-mnemonic commodity))))
       (and
        commodity
@@ -395,23 +420,23 @@
          (currency-cmd-list (call-with-values 
                              (lambda () (partition!
                                          (lambda (cmd)
-                                           (not (eq? (car cmd) 'currency)))
+                                           (not (equal? (car cmd) "currency")))
                                          big-list))
                              (lambda (a b) (set! cmd-list a) b)))
          (cmd-hash (make-hash-table 31)))
 
     ;; Now collect symbols going to the same backend.
-    (item-list->hash! cmd-list cmd-hash car cdr hashq-ref hashq-set! #t)
+    (item-list->hash! cmd-list cmd-hash car cdr hash-ref hash-set! #t)
 
     ;; Now translate to just what finance-quote-helper expects.
     (append
      (hash-fold
       (lambda (key value prior-result)
-        (cons (cons (fq-method-sym->str key) value)
+        (cons (cons key value)
               prior-result))
       '()
       cmd-hash)
-     (map (lambda (cmd) (cons (fq-method-sym->str (car cmd)) (list (cdr cmd))))
+     (map (lambda (cmd) (cons (car cmd) (list (cdr cmd))))
           currency-cmd-list))))
 
   (define (fq-call-data->fq-calls fq-call-data)
@@ -562,7 +587,7 @@
       (if (not (and commodity currency gnc-time price price-type))
           (string-append
            currency-str ":" (gnc:commodity-get-mnemonic commodity))
-          (let ((gnc-price (gnc:price-create)))
+          (let ((gnc-price (gnc:price-create book)))
             (if (not gnc-price)
                 (string-append
                  currency-str ":" (gnc:commodity-get-mnemonic commodity))
@@ -594,7 +619,7 @@
                                (map fq-call-data->fq-calls fq-call-data))))
          (fq-results (and fq-calls (gnc:fq-get-quotes fq-calls)))
          (commod-tz-quote-triples
-          (and fq-results (not (member 'missing-lib fq-results))
+          (and fq-results (list? (car fq-results))
                (fq-results->commod-tz-quote-triples fq-call-data fq-results)))
          ;; At this point commod-tz-quote-triples will either be #f or a
          ;; list of items. Each item will either be (commodity
@@ -616,6 +641,16 @@
          (keep-going? #t))
 
     (cond
+     ((eq? quotables #f)
+      (set! keep-going? #f)
+      (if (gnc:ui-is-running?)
+          (gnc:error-dialog  (_ "No accounts marked for quote retrieval."))
+	  (gnc:warn (_ "No accounts marked for quote retrieval."))))
+     ((eq? fq-results #f)
+      (set! keep-going? #f)
+      (if (gnc:ui-is-running?)
+          (gnc:error-dialog  (_ "Unable to get quotes or diagnose the problem."))
+	  (gnc:warn (_ "Unable to get quotes or diagnose the problem."))))
      ((member 'missing-lib fq-results)
       (set! keep-going? #f)
       (if (gnc:ui-is-running?)
@@ -624,6 +659,18 @@
 Run 'update-finance-quote' as root to install them."))
           (gnc:warn (_ "You are missing some needed Perl libraries.
 Run 'update-finance-quote' as root to install them.") "\n")))
+     ((member 'system-error fq-results)
+      (set! keep-going? #f)
+      (if (gnc:ui-is-running?)
+          (gnc:error-dialog
+           (_ "There was a system error while retrieving the price quotes."))
+          (gnc:warn (_ "There was a system error while retrieving the price quotes.") "\n")))
+     ((not (list? (car fq-results)))
+      (set! keep-going? #f)
+      (if (gnc:ui-is-running?)
+          (gnc:error-dialog
+           (_ "There was an unknown error while retrieving the price quotes."))
+          (gnc:warn (_ "There was an unknown error while retrieving the price quotes.") "\n")))
      ((and (not commod-tz-quote-triples) (gnc:ui-is-running?))
       (gnc:error-dialog
        (_ "Unable to get quotes or diagnose the problem."))
@@ -636,7 +683,7 @@ Run 'update-finance-quote' as root to install them.") "\n")))
           (if (and ok-syms (not (null? ok-syms)))
               (set!
                keep-going?
-               (gnc:verify-dialog
+               (gnc:verify-dialog #t
                 (call-with-output-string
                  (lambda (p)
                    (display (_ "Unable to retrieve quotes for these items:") p)
@@ -644,8 +691,7 @@ Run 'update-finance-quote' as root to install them.") "\n")))
                    (display "  " p)
                    (display (string-join problem-syms "\n  ") p)
                    (newline p)
-                   (display (_ "Continue using only the good quotes?") p)))
-                #t))
+                   (display (_ "Continue using only the good quotes?") p)))))
               (begin
                 (gnc:error-dialog
                  (call-with-output-string
@@ -676,7 +722,7 @@ Run 'update-finance-quote' as root to install them.") "\n")))
            (if (gnc:ui-is-running?)
                (set!
                 keep-going?
-                (gnc:verify-dialog
+                (gnc:verify-dialog #t
                  (call-with-output-string
                   (lambda (p)
                     (display (_ "Unable to create prices for these items:") p)
@@ -684,8 +730,7 @@ Run 'update-finance-quote' as root to install them.") "\n")))
                     (display "  " p)
                     (display (string-join (filter string? prices) "\n  ") p)
                     (newline p)
-                    (display (_ "Add remaining good quotes?") p)))
-                 #t))
+                    (display (_ "Add remaining good quotes?") p)))))
                (gnc:warn
                 (call-with-output-string
                  (lambda (p)
@@ -703,18 +748,23 @@ Run 'update-finance-quote' as root to install them.") "\n")))
                                    prices)))))))
 
 (define (gnc:add-quotes-to-book-at-url url)
-  (let* ((book (gnc:url->loaded-book url #f #f))
-         (quote-ok? (and book (gnc:book-add-quotes book))))
+  (let* ((session (gnc:url->loaded-session url #f #f))
+         (quote-ok? #f))
+    (if session
+	(begin
+	  (set! quote-ok? (and (gnc:book-add-quotes
+				(gnc:session-get-book session))))
 
-    (if (not quote-ok?) (gnc:msg "book-add-quotes failed"))
-    (and book (gnc:book-save book))
-    (if (not (eq? 'no-err
-                  (gw:enum-<gnc:BackendError>-val->sym
-                   (gnc:book-get-error book) #f)))
-        (set! quote-ok? #f))
-    (if (not quote-ok?)
-        (gnc:msg "book-save failed " (gnc:book-get-error book)))
-    (and book (gnc:book-destroy book))
+	  (if (not quote-ok?) (gnc:msg "book-add-quotes failed"))
+	  (gnc:session-save session)
+	  (if (not (eq? 'no-err
+			(gw:enum-<gnc:BackendError>-val->sym
+			 (gnc:session-get-error session) #f)))
+	      (set! quote-ok? #f))
+	  (if (not quote-ok?)
+	      (gnc:msg "session-save failed " (gnc:session-get-error session)))
+	  (gnc:session-destroy session))
+	(gnc:error "book-add-quotes unable to open file"))
     quote-ok?))
 
 ; (define (get-1-quote exchange . items)
